import numpy as np
import os

class Parameters:

	def __init__(self):
		'''
		default parameters

		e.g. a 50 MWe power plant (Yogi Goswami, Principles of Solar Engineering, Third Edition, page 480, section 8.8)
			 design point DNI 950 W/m2, SM=1.8
			 the required field rating is 305 MW
			 the total reflector area is approximated as 483,000 m2
			 tower height 137 m for a surround field, 183 m for a polar field

			for PS10, the annual capacity of the field is around 120GWh
		'''
		self.simulation()
		self.Sun()
		self.Heliostat()
		self.Receiver()
		self.dependent_par()

	def Sun(self):
		'''
			(1) lat       : float, latitude of the plant lcation
			(2) dni_des       : float, dni at design point (W/m2)   
			(3) sunshape  :   str, 'pillbox' or 'Buie' 
			(4) sunsize   : float, 
				-- for 'pillbox': half angle of the pillbox sunshape (degree)
				-- for 'Buie': circumsolar ratio (CSR)  

			(5) extinction: float, around 1e-6, the extinction coefficient if the atmosphere is surrounded by participant medium
		'''
		self.lat=37.44 # latitude, default: location of PS10
		self.dni_des=1000.
		self.sunshape='pillbox'
		self.crs=0.01
		self.half_angle_deg=4.65*1.e-3*180./np.pi # convert rad to degree --> solstice convention
		self.std_dev=0.1
		self.extinction=1e-6
		self.wea_file=None


	def Heliostat(self):
		'''
			(1) field_type : str,
				-- 'polar', 'polar-half', 'surround' or 'surround-half' for desiging a new field 
				-- the directory of the layout file
				    the layout file is a 'csv' file, (n+2, 7)
				   - n is the total number of heliostats 
				   - the 1st row is the number of each column
				   - the 2nd row is the unit 
				   - the first three columns are X, Y, Z coordinates of each heliostat
				   - the fourth column is the focal length
				   - the last three columns are the aiming points
			(2) Q_in_rcv   : float, required heat of the receiver (W)
			(3) W_helio    : float, width of a heliostat (m) 
			(4) H_helio    : float, height of a heliostat (m)
			(5) slope_error: float, slope error(radians)
			(6) rho_helio  : float, reflectivity of heliostat 
			(7) H_tower    : float, tower height (m)
			(8) R_tower    : float, radius of tower (m)
			(9) concret_tower: bool, True-a solid concrete tower, or False-a truss tower
			(10)single_filed : bool, True-one tower one field, or False-multi-tower
			(11)R1         : float, layout parameter, the distance of the first row of heliostat 
			(12)dsep       : float, layout parameter, the separation distance of heliostats (m)
			(13)fb         : float, in (0-1), a factor to expand the field to reduce block 
			---(*) n_helios:   int, number of heliostats for the designed field 
			---(*) Z_helio : float, the installation height of the heliostat
		'''

		self.field_type='polar'
		if self.method==1:
			self.Q_in_rcv=10e6 # required heat of the receiver  
		else:
			self.n_helios=1000
		self.W_helio=10.
		self.H_helio=10.
		self.slope_error=2.e-3 # radian
		self.rho_helio=0.9
		self.H_tower=100.
		self.R_tower=0.001  # shading effect of tower is neglected at the moment
		self.concret_tower=False
		self.single_field=True 

		self.R1=90.
		self.dsep=0.
		self.fb=0.7

	def Receiver(self):
		'''
			(1) rcv_type  :   str, 'flat', 'cylinder', 'particle' or 'stl', type of the receiver
			(2) H_rcv     : float, height of a flat receiver or radius of a cylindrical receiver (m) 
			(3) W_rcv     : float, width of the receiver (m)
			(4) tilt_rcv  : float, tilt angle of the receiver (deg), 0 is where the receiver face to the horizontal
			(5) alpha_rcv : float, receiver surface absorptivity (0-1), set as 1 if rcv_type='particle'
			(6) n_H_rcv   :   int, number of discretisation of the receiver in the height direction
			(7) n_W_rcv   :   int, number of discretisation of the receiver in the width direction (for a flat receiver) 
				                    or in the circular direction (for a cylindrical receiver) 
			(8) X_rcv     : float, x location of the receiver (m)
			(9) Y_rcv     : float, y location of the receiver (m)
			---(*) z_rcv  : float, z location of the receiver (m)

		'''
		self.rcv_type='flat'
		self.H_rcv=10.
		self.W_rcv=10.
		self.tilt_rcv=0. # receiver tilt angle
		self.alpha_rcv=1. 
		self.n_H_rcv=10
		self.n_W_rcv=10
		self.X_rcv=0. # receiver location
		self.Y_rcv=0.


	def simulation(self):
		'''
		n_row_oelt: int, number of rows of the lookup table (i.e. simulated days per year)
		n_col_oelt: int, number of columns of the lookup table (i.e. simulated number of hours per day)
		n_rays    : int, number of rays for the simulation
		n_procs   : int, number of processors for the mcrt simulation 
		casedir   : str, the directory for saving the result files
		'''  
		self.n_row_oelt=5
		self.n_col_oelt=5
		self.n_rays=int(5e6)
		self.n_procs=1
		self.casedir='.'
		self.method=1 # 1 - design the field based on the Q_in_rcv
				      # 2 - design the field based on the n_helios
    

	def dependent_par(self):
		'''
		'''
		self.Z_helio=self.H_helio*0.7
		self.Z_rcv=self.H_tower        
		if self.lat>=0:
			self.hemisphere='North'
		elif self.lat<0:
			self.hemisphere='South'

		# estimate a rough number of large field
		eta_field=0.4 # assumed field effieicy at design point
		if self.method==1:
			self.n_helios=self.Q_in_rcv/self.W_helio/self.H_helio/self.dni_des/eta_field
			  
			if self.field_type=='polar':
				self.n_helios*=2.  

	def saveparam(self, savedir):
		if not os.path.exists(savedir):
			os.makedirs(savedir)
    
		param=np.array([
				['method', self.method, '-'],    
				['field', self.field_type, '-'],  
				['Q_in_rcv', self.Q_in_rcv, 'W'],     
				['n_helios(pre_des if method ==1)', self.n_helios, '-'],    
				['W_helio', self.W_helio, 'm'],    
				['H_helio', self.H_helio, 'm'],
				['Z_helio', self.Z_helio, 'm'],      
				['rho_helio', self.rho_helio, '-'],    
				['slope_error', self.slope_error, 'rad'],
				['H_tower', self.H_tower, 'm'],    
				['R_tower', self.R_tower, 'm'], 
				['concret_tower', self.concret_tower, '-'],    
				['single_field', self.single_field, '-'],  
				['fb factor', self.fb, '-'],     
				['R1', self.R1, 'm'],    
				['dsep', self.dsep, 'm'],    
				['rcv_type', self.rcv_type, '-'],  
				['H_rcv', self.H_rcv, 'm'],
				['W_rcv', self.W_rcv, 'm'],   
				['tilt_rcv', self.tilt_rcv, 'deg'],  
				['alpha_rcv', self.alpha_rcv, '-'],  
				['n_H_rcv', self.n_H_rcv, '-'],   
				['n_W_rcv', self.n_W_rcv, '-'],    
				['X_rcv', self.X_rcv, 'm'],   
				['Y_rcv', self.Y_rcv, 'm'],   
				['Z_rcv', self.Z_rcv, 'm'],   
				['n_row_oelt', self.n_row_oelt, '-'] , 
				['n_col_oelt', self.n_col_oelt, '-'] ,
				['n_rays', self.n_rays, '-'] ,
				['n_procs', self.n_procs, '-'] 
				])

		np.savetxt(savedir+'/simulated_parameters.csv', param, delimiter=',', fmt='%s')
    




import numpy as np
#from tracer.models.heliostat_field import solar_vector
#import matplotlib.pyplot as plt
from scipy.spatial import Delaunay
from .gen_vtk import gen_vtk

class FieldPF:
	"""Preliminary calculation of heliostat field performance.

	1. cosine factors: sun - heliostat
	2. another cosine factors: receiver - heliostat

	Note the angle conventions in this program:

	  * azimuth: the solar azimuth angle, from South to West
	  * zenith: the solar zenith angle, 0 from vertical


	``Example``

		>>> from solsticepy.cal_field import *
		>>> pos_and_aiming=np.loadtxt('./pos_and_aiming.csv', delimiter=',', skiprows=2) #load the field layout file (refer to cal_layout.py)
		>>> pos=pos_and_aiming[:,:3]
		>>> aim=pos_and_aiming[:,4:]
		>>> azimuth=np.r_[0.]
		>>> zenith=np.r_[12.]
		>>> field=FieldPF(np.r_[0,1,0])
		>>> sun_vec=field.get_solar_vector(azimuth, zenith)
		>>> norms=field.get_normals(towerheight=70., hstpos=pos, sun_vec=sun_vec)
		>>> COORD, TRI, ele, nc=field.mesh_heliostat_field(width=10., height=8., normals=norms, hstpos=pos)
		>>> cos=field.get_cosine(hst_norms=norms, sun_vec=sun_vec)
		>>> savedir='./field.vtk'
		>>> COS=np.repeat(cos, ele)
		>>> DATA={'cos':COS}
		>>> NORMS=np.repeat(norms, ele, axis=0)
		>>> gen_vtk(savedir, COORD.T, TRI, NORMS, True, DATA)

		The field.vtk file is saved in the local directory and can be open in ParaView to visualise the cosine factor of each individual heliostat. See more details in the 'gen_vtk.py' in the next section.


	"""

	def __init__(self, receiver_norm=np.r_[0,1,0]):
		"""

		``Argument``

		  * receiver_normal (numpy array): A numpy 3-vector with unit normal direction of the receiver aperature (default: [0,1,0])

		"""
		self.rec_norm=receiver_norm.reshape(3,1)


	def get_solar_vector(self, azimuth, zenith):
		"""Calculate the solar vector using azimuth and zenith angles.

		``Arguments``

		  * azimuth (float): the sun's azimuth (deg), from South increasing towards to the West
		  * zenith (float): angle created between the solar vector and the Z axis (deg)

		``Return``

		  * sun_vec (numpy array): a 3-component 1D array with the solar vector
		"""
		#copied from tracer.models.heliostat_field import solar_vector
		#TODO change it to Solstice convetion if necessary

		azimuth*=np.pi/180.
		zenith*=np.pi/180.

		sun_z = np.cos(zenith)
		sun_y=-np.sin(zenith)*np.cos(azimuth)
		sun_x=-np.sin(zenith)*np.sin(azimuth)
		sun_vec = np.r_[sun_x, sun_y,sun_z] 

		return sun_vec

	def get_normals(self, towerheight, hstpos, sun_vec):
		"""Calculate the normal vectors of each heliostat

		``Arguments``

		  * towerheight (float): tower height
		  * hstpos (nx3 numpy array): heliostat positions
		  * sun_vec (numpy array): the solar vector

		``Return``

		  * hst_norms (numpy array): normal vectors of each heliostat

		"""

		tower_vec=-hstpos
		tower_vec[:,-1]+=towerheight
		tower_vec/=np.sqrt(np.sum(tower_vec**2, axis=1)[:,None])

		hst_norms=sun_vec+tower_vec
		hst_norms/=np.sqrt(np.sum(hst_norms**2, axis=1)[:,None])        
		return hst_norms

	def get_rec_view(self, towerheight, hstpos):
		"""Check the visibility of each heliostat from the view of the receiver
	
		``Arguments``
	
		  * towerheight (float): tower height 
		  * hstpos (numpy array): position of each heliostat

		``Return``

		  * vis_idx (numpy array): the indices of the heliostats that can be seen by the receiver 
		
		"""

		# angle between normal of the receiver apterture and the RH
		# RH is the vector from the receiver to the heliostat (i.e. -tower_vec)

		tower_vec=-hstpos
		tower_vec[:,-1]+=towerheight
		tower_vec/=np.sqrt(np.sum(tower_vec**2, axis=1)[:,None])

		view=np.arccos(np.dot(-tower_vec, self.rec_norm))
		view=view.flatten()
		vis_idx=(view<1.5) # the heliostat that can be seen by the receiver

		return vis_idx


	def get_cosine(self,hst_norms, sun_vec):

		"""Calculate the cosine factor between the sun and the heliostats

		``Arguments``

		  * hst_norms (numpy array): normal vectors of the heliostats
		  * sun_vec (numpy array): solar vector

		``Return``

		  * cos_factor (numpy array): the cosine factor between the sun and the heliostats
		
		"""

		cos_factor=np.sum(hst_norms*sun_vec, axis=1)

		return cos_factor

	def mesh_heliostat(self, width, height):
		"""The local coordinate of the triangular mesh of a heliostat 

		``Arguments``
		  * width (float): width of the heliostat
		  * height (float): height of the heliostat

		``Returns``

		  * coords (numpy array): coordinates of the vertices
		  * tri (numpy array): the indices of the triangular mesh

		"""
		x=np.linspace(-width/2., width/2., 2)
		y=np.linspace(-height/2., height/2., 2)

		xx, yy=np.meshgrid(x, y)
		coords=np.column_stack([xx.ravel(),yy.ravel()])   
		tri=Delaunay(coords).simplices
		#plt.figure(1) 
		#plt.triplot(coords[:,0], coords[:,1], tri)   
		#plt.show()
		return coords, tri

	def mesh_heliostat_field(self, width, height, normals, hstpos):
		"""Generate the necessary elements to create the VTK file to view the heliostat layout 

		``Arguments``
		  * width (float): width of the heliostat
		  * height (float): height of the heliostat
		  * normals (numpy array): normal vectors of the heliostats
		  * hstpos (numpy array): the positions of the heliostats

		``Returns``

		  * COORD (numpy array): the coordinates of the indices of the helisotat field
		  * TRI (numpy array): the indices of the triangular mesh of the heliostat field
		  * ele (int): number of element of the triangular mesh
		  * nc (int): number of indices
		"""		


		coord1, tri1=self.mesh_heliostat(width, height)

		ele=len(tri1)
		nc=len(coord1)

		num_hst=len(hstpos)

		COORD=np.zeros(num_hst*nc*3).reshape(num_hst*nc, 3)
		TRI=np.zeros(num_hst*ele*3).reshape(num_hst*ele, 3)

		norm_x=normals[:,0]
		norm_y=normals[:,1]
		norm_z=normals[:,2]          

		for i in range(num_hst):            
		 
		    TRI[i*ele: (i+1)*ele]=tri1+i*nc

		    trans = np.eye(4)
		    trans[:3,3] = hstpos[i]

		    x=coord1[:,0]
		    y=coord1[:,1]
		    z=float(hstpos[i,2])*np.ones(nc)  

		    cd=np.vstack((x, y, z, np.ones(nc)))   
		    cd_t=np.dot(trans, cd)

		    xx=cd_t[0]
		    yy=cd_t[1]
		    zz=cd_t[2]

		    COORD[i*nc: (i+1)*nc,0]=xx
		    COORD[i*nc: (i+1)*nc,1]=yy
		    COORD[i*nc: (i+1)*nc,2]=zz

		#plt.figure(1) 
		#plt.triplot(COORD[:,0], COORD[:,1], TRI)   
		#plt.show()        
		return COORD, TRI, ele, nc

	def plot_cosine(self, savename):
		"""Plot the cosine factors of heliostats in Matplotlib

		``Argument``

		  * savename (str): the directory to save the figure, with suffix, e.g. '.png' or '.jpg'

		``Return``

		  * No return value (a figure is written in the specified path)

		"""
		x=self.hstpos[:,0]
		y=self.hstpos[:,1]
		z=self.hstpos[:,2]
		av=(self.view<np.pi/2.)
		plt.figure(1)
		cm = plt.cm.get_cmap('rainbow')
		cs=plt.scatter(x[av], y[av], c=self.cosine_factor[av], cmap=cm,s=30)
		plt.colorbar(cs)
		plt.title('Cosine factors')
		plt.savefig(open(savename, 'w'),dpi=500, bbox_inches='tight')
		plt.close()	


	def plot_select(self, savename, tilt):
		"""Plot the heliostats that can be seen by the receiver in Matplotlib

		``Argument``

		  * savename (str): the directory to save the figure, with suffix, e.g. '.png' or '.jpg'
		  * savename (float): receiver tilted angle

		``Return``

		  * No return value (a figure is written in the specified path)

		"""

		x=self.hstpos[:,0]
		y=self.hstpos[:,1]
		z=self.hstpos[:,2]

		plt.figure(1)
		fig,ax1=plt.subplots()
		cm = plt.cm.get_cmap('rainbow')
	 
		av=(self.view<1.5)
		cs=plt.scatter(x[av], y[av], c=self.view[av], cmap=cm)
		plt.colorbar(cs)
		nv=(self.view>=1.5)
		plt.scatter(x[nv], y[nv], c='gray')
		plt.title('Receiver view \n tilt %s deg'%pm )
		#plt.title('Receiver view \n Tower height %s m'%pm )
		plt.savefig(open(savename, 'w'),dpi=500, bbox_inches='tight')
		plt.close()	
       
def rotx(ang):
    """Generate a homogenous transform for ang radians around the x axis"""
    s = np.sin(ang); c = np.cos(ang)
    return np.array([
        [1., 0, 0, 0],
        [0, c,-s, 0],
        [0, s, c, 0],
        [0, 0, 0, 1.]
    ])

def roty(ang):
    """Generate a homogenous transform for ang radians around the y axis"""
    s = np.sin(ang); c = np.cos(ang)
    return np.array([
        [c, 0, s, 0],
        [0, 1., 0, 0],
        [-s,0, c, 0],
        [0, 0, 0, 1.]
    ])

def rotz(ang):
    """Generate a homogenous trransform for ang radians around the z axis"""
    s = np.sin(ang); c = np.cos(ang)
    return np.array([
        [c,-s, 0, 0],
        [s, c, 0, 0],
        [0, 0, 1., 0],
        [0, 0, 0, 1.]
    ])

def translate(x=0, y=0, z=0):
    """Generate a homogenous transform for translation by x, y, z"""
    return np.array([
        [1., 0, 0, x],
        [0, 1., 0, y],
        [0 ,0, 1., z],
        [0, 0, 0, 1.]
    ])
          

if __name__=='__main__':
    #pos_and_aiming=np.loadtxt('/media/yewang/Work/svn_ye/Solstice-tutorial/cases/2-Validation/layout.csv', delimiter=',', skiprows=2)
    pos_and_aiming=np.loadtxt('./pos_and_aiming.csv', delimiter=',', skiprows=2)
    pos=pos_and_aiming[:,:3]
    aim=pos_and_aiming[:,4:]
    azimuth=np.r_[0.]
    zenith=np.r_[12.]
    field=FieldPF(np.r_[0,1,0])
    sun_vec=field.get_solar_vector(azimuth, zenith)
    norms=field.get_normals(towerheight=70., hstpos=pos, sun_vec=sun_vec)
    #field.heliostat(10, 8)
    COORD, TRI, ele, nc=field.view_heliostats(width=10., height=8., normals=norms, hstpos=pos)
    cos=field.get_cosine(hst_norms=norms, sun_vec=sun_vec)
    savedir='./field.vtk'
    COS=np.repeat(cos, ele)
    DATA={'cos':COS}
    NORMS=np.repeat(norms, ele, axis=0)
    gen_vtk(savedir, COORD.T, TRI, NORMS, True, DATA)
        
